from __future__ import annotations

import copy
import hashlib
import os
import shutil
from pathlib import Path
from typing import TYPE_CHECKING, Any, Generator, Literal

import aiofiles
import ultima_scraper_api
from aiohttp import ClientResponse
from aiohttp.client_reqrep import ClientResponse
from ultima_scraper_api.helpers.main_helper import open_partial
from ultima_scraper_api.managers.session_manager import EXCEPTION_TEMPLATE
from ultima_scraper_collection.helpers import main_helper as usc_helper
from ultima_scraper_renamer.reformat import (
    FormatAttributes,
    ReformatItem,
    ReformatManager,
)

if TYPE_CHECKING:
    api_types = ultima_scraper_api.api_types
    user_types = ultima_scraper_api.user_types
    from ultima_scraper_collection import datascraper_types
    from ultima_scraper_collection.config import site_config_types


class FilesystemManager:
    def __init__(self) -> None:
        self.user_data_directory = Path("__user_data__")
        self.trash_directory = self.user_data_directory.joinpath("trash")
        self.profiles_directory = self.user_data_directory.joinpath("profiles")
        self.devices_directory = self.user_data_directory.joinpath("drm_device")
        self.settings_directory = Path("__user_data__")
        self.ignore_files = ["desktop.ini", ".DS_Store", ".DS_store", "@eaDir"]
        self.directory_manager: DirectoryManager | None = None
        self.directory_manager_users: dict[int, DirectoryManager] = {}
        self.file_manager_users: dict[int, FileManager] = {}

    def __iter__(self):
        for each in self.__dict__.values():
            yield each

    def check(self):
        for directory in self:
            if isinstance(directory, Path):
                directory.mkdir(exist_ok=True)

    def move(self, src: Path, trg: Path):
        shutil.move(src, trg)

    def remove_mandatory_files(
        self, files: list[Path] | Generator[Path, None, None], keep: list[str] = []
    ):
        folders = [x for x in files if x.name not in self.ignore_files]
        if keep:
            folders = [x for x in files if x.name in keep]
        return folders

    def get_directory_manager(self, user_id: int):
        return self.directory_manager_users[user_id]

    def get_file_manager(self, user_id: int):
        return self.file_manager_users[user_id]

    def activate_directory_manager(self, site_config: site_config_types):
        root_metadata_directory = usc_helper.check_space(
            site_config.metadata_setup.directories
        )
        root_download_directory = usc_helper.check_space(
            site_config.download_setup.directories
        )
        self.directory_manager = DirectoryManager(
            site_config,
            root_metadata_directory,
            root_download_directory,
        )

    def trash(self):
        pass

    async def write_data(
        self,
        response: ClientResponse,
        download_path: Path,
        callback: Any = None,
        buffer_size: int = 64 * 1024 * 1024,  # 512MB buffer
    ):
        status_code = None
        if response.status == 200:
            total_length = 0
            os.makedirs(os.path.dirname(download_path), exist_ok=True)

            async with aiofiles.open(f"{download_path}.part", mode="wb") as f:
                buffer = bytearray()  # Buffer to hold chunks before writing to disk
                try:
                    async for data in response.content.iter_chunked(
                        65536
                    ):  # 64KB chunks from network
                        buffer.extend(data)  # Add chunk to buffer
                        length = len(data)
                        total_length += length

                        if len(buffer) >= buffer_size:
                            await f.write(buffer)
                            buffer.clear()  # Clear buffer after writing

                        if callback:
                            callback(length)
                except EXCEPTION_TEMPLATE as _e:
                    status_code = 1
                except Exception as _e:
                    raise Exception(f"Unknown Error: {_e}")
                else:
                    if status_code:
                        await aiofiles.os.remove(f"{download_path}.part")
                    else:
                        try:
                            if buffer:  # Write any remaining buffered data
                                await f.write(buffer)
                            await aiofiles.os.rename(
                                f"{download_path}.part", download_path
                            )
                        except OSError as _e:
                            raise Exception(f"File rename failed: {_e}")
        else:
            if response.content_length:
                pass
            status_code = 2
        return status_code

    async def create_option(
        self,
        datascraper: datascraper_types,
        username: str,
        directory: Path,
        format_key: str,
    ):
        api = datascraper.api
        option = {
            "site_name": api.site_name,
            "profile_username": username,
            "model_username": username,
            "directory": directory,
        }
        reformat_item_fd = ReformatItem(option)
        assert self.directory_manager
        f_d_p = reformat_item_fd.remove_non_unique(self.directory_manager, format_key)
        return f_d_p

    async def create_directory_manager(
        self, site_config: site_config_types, user: user_types
    ):
        if self.directory_manager:
            final_download_directory = await self.discover_main_directory(user)
            final_root_download_directory = (
                self.directory_manager.root_download_directory
            )
            for directory in site_config.download_setup.directories:
                assert directory.path
                if directory.path.as_posix() in final_download_directory.as_posix():
                    final_root_download_directory = directory.path
                    break
            directory_manager = DirectoryManager(
                site_config,
                self.directory_manager.root_metadata_directory,
                final_root_download_directory,
            )
            self.directory_manager_users[user.id] = directory_manager
            self.file_manager_users[user.id] = FileManager(directory_manager)
            return directory_manager

    async def discover_main_metadata_directory(self, subscription: user_types):
        usernames = subscription.get_usernames(ignore_id=False)
        valid_usernames = subscription.get_usernames(ignore_id=True)
        authed = subscription.get_authed()
        reformat_manager = ReformatManager(authed, self)
        directory_manager = self.directory_manager
        site_config = directory_manager.site_config
        final_store_directory = None
        for username in usernames:
            for store_directory in [
                x.path for x in site_config.metadata_setup.directories if x.path
            ]:
                download_directory_reformat_item = (
                    reformat_manager.prepare_user_reformat(
                        subscription, store_directory, username=username
                    )
                )
                formatted_download_directory = (
                    download_directory_reformat_item.reformat(
                        site_config.metadata_setup.directory_format
                    )
                )
                final_store_directory = formatted_download_directory
                if final_store_directory.exists():
                    if username == f"u{subscription.id}":
                        if valid_usernames:
                            download_directory_reformat_item = (
                                reformat_manager.prepare_user_reformat(
                                    subscription,
                                    store_directory,
                                    username=valid_usernames[-1],
                                )
                            )
                            formatted_download_directory = (
                                download_directory_reformat_item.reformat(
                                    site_config.metadata_setup.directory_format
                                )
                            )
                            if not formatted_download_directory.exists():
                                formatted_download_directory.mkdir(
                                    exist_ok=True, parents=True
                                )
                                final_store_directory.rename(
                                    formatted_download_directory
                                )
                                final_store_directory = formatted_download_directory
                            else:
                                final_store_directory = formatted_download_directory
                            return final_store_directory
                    else:
                        return final_store_directory
        return final_store_directory

    async def discover_main_directory(self, subscription: user_types):
        usernames = subscription.get_usernames(ignore_id=False)
        if f"u{subscription.id}" not in usernames:
            usernames.append(f"u{subscription.id}")
        valid_usernames = subscription.get_usernames(ignore_id=True)
        authed = subscription.get_authed()
        reformat_manager = ReformatManager(authed, self)
        directory_manager = self.directory_manager
        assert directory_manager
        site_config = directory_manager.site_config
        store_directories = [
            x.path for x in site_config.download_setup.directories if x.path
        ]

        for username in usernames:
            for store_directory in store_directories:
                download_directory_reformat_item = (
                    reformat_manager.prepare_user_reformat(
                        subscription, store_directory, username=username
                    )
                )
                formatted_download_directory = (
                    download_directory_reformat_item.remove_non_unique(
                        directory_manager, "file_directory_format"
                    )
                )

                if formatted_download_directory.exists():
                    if username == f"u{subscription.id}" and valid_usernames:
                        download_directory_reformat_item = (
                            reformat_manager.prepare_user_reformat(
                                subscription,
                                store_directory,
                                username=valid_usernames[-1],
                            )
                        )
                        new_formatted_download_directory = (
                            download_directory_reformat_item.remove_non_unique(
                                directory_manager, "file_directory_format"
                            )
                        )
                        if not new_formatted_download_directory.exists():
                            # formatted_download_directory.mkdir(
                            #     exist_ok=True, parents=True
                            # )
                            formatted_download_directory.rename(
                                new_formatted_download_directory
                            )
                            formatted_download_directory = (
                                new_formatted_download_directory
                            )
                        return formatted_download_directory
                    return formatted_download_directory

        download_directory_reformat_item = reformat_manager.prepare_user_reformat(
            subscription, directory_manager.root_download_directory, username=username
        )
        formatted_download_directory = (
            download_directory_reformat_item.remove_non_unique(
                directory_manager, "file_directory_format"
            )
        )
        return formatted_download_directory

    async def discover_alternative_directories(self, subscription: user_types):
        usernames = subscription.get_usernames(ignore_id=False)
        authed = subscription.get_authed()
        reformat_manager = ReformatManager(authed, self)
        directory_manager = self.get_directory_manager(subscription.id)
        site_config = directory_manager.site_config
        for username in usernames:
            for alt_download_directory in [
                x.path for x in site_config.download_setup.directories if x.path
            ]:
                alt_download_directory_reformat_item = (
                    reformat_manager.prepare_user_reformat(
                        subscription, alt_download_directory, username=username
                    )
                )
                formatted_alt_download_directory = (
                    alt_download_directory_reformat_item.remove_non_unique(
                        directory_manager, "file_directory_format"
                    )
                )
                if (
                    formatted_alt_download_directory
                    == directory_manager.user.download_directory
                ):
                    continue
                if formatted_alt_download_directory.exists():
                    directory_manager.user.alt_download_directories.append(
                        formatted_alt_download_directory
                    )
        return directory_manager.user.alt_download_directories

    async def format_directories(self, performer: user_types) -> DirectoryManager:
        authed = performer.get_authed()
        directory_manager = self.get_directory_manager(performer.id)
        file_manager = self.get_file_manager(performer.id)

        final_metadata_directory = await self.discover_main_metadata_directory(
            performer
        )
        directory_manager.user.metadata_directory = final_metadata_directory

        final_download_directory = await self.discover_main_directory(performer)
        directory_manager.user.download_directory = final_download_directory

        api = authed.api
        performer_username = performer.get_usernames(ignore_id=True)[-1]
        site_name = authed.api.site_name
        alt_directories = await self.discover_alternative_directories(performer)
        await file_manager.set_default_files()
        _metadata_filepaths = await file_manager.find_metadata_files(legacy_files=False)
        # for metadata_filepath in metadata_filepaths:
        #     if file_manager.directory_manager.user.metadata_directory.as_posix() in metadata_filepath.parent.as_posix():
        #         continue
        #     new_filepath = file_manager.directory_manager.user.metadata_directory.joinpath(metadata_filepath.name)
        #     if new_filepath.exists():
        #         new_filepath = usc_helper.find_unused_filename(
        #             new_filepath
        #         )
        #         if new_filepath.exists():
        #             breakpoint()
        #     file_manager.rename_path(metadata_filepath, new_filepath)
        #     pass
        # alt_files = await usc_helper.walk(
        #     file_manager.directory_manager.user.download_directory
        # )
        # if not alt_files:
        #     shutil.rmtree(file_manager.directory_manager.user.download_directory)
        await file_manager.merge_alternative_directories(alt_directories)
        user_metadata_directory = directory_manager.user.metadata_directory
        assert user_metadata_directory
        _user_download_directory = directory_manager.user.download_directory
        legacy_metadata_directory = user_metadata_directory
        directory_manager.user.legacy_metadata_directories.append(
            legacy_metadata_directory
        )
        items = api.CategorizedContent()
        for api_type, _ in items:
            legacy_metadata_directory_2 = user_metadata_directory.joinpath(api_type)
            directory_manager.user.legacy_metadata_directories.append(
                legacy_metadata_directory_2
            )
        legacy_model_directory = directory_manager.root_download_directory.joinpath(
            site_name, performer_username
        )
        directory_manager.user.legacy_download_directories.append(
            legacy_model_directory
        )
        return directory_manager


class DirectoryManager:
    def __init__(
        self,
        site_config: site_config_types,
        root_metadata_directory: Path,
        root_download_directory: Path,
    ) -> None:
        self.root_directory = Path()
        self.root_metadata_directory = Path(root_metadata_directory)
        self.root_download_directory = Path(root_download_directory)
        self.user = self.UserDirectories()
        self.site_config = site_config
        formats = FormatTypes(site_config)
        string, status = formats.check_rules()
        if not status:
            print(string)
            exit(0)
        self.formats = formats
        pass

    def create_directories(self):
        # self.profile.create_directories()
        self.root_metadata_directory.mkdir(exist_ok=True)
        self.root_download_directory.mkdir(exist_ok=True)

    def delete_empty_directories(
        self, directory: Path, filesystem_manager: FilesystemManager
    ):
        for root, dirnames, _files in os.walk(directory, topdown=False):
            for dirname in dirnames:
                full_path = os.path.realpath(os.path.join(root, dirname))
                contents = os.listdir(full_path)
                if not contents:
                    shutil.rmtree(full_path, ignore_errors=True)
                else:
                    content_paths = [Path(full_path, content) for content in contents]
                    contents = filesystem_manager.remove_mandatory_files(content_paths)
                    if not contents:
                        shutil.rmtree(full_path, ignore_errors=True)

        if os.path.exists(directory) and not os.listdir(directory):
            os.rmdir(directory)

    # class ProfileDirectories:
    #     def __init__(self, root_directory: Path) -> None:
    #         self.root_directory = Path(root_directory)
    #         self.metadata_directory = self.root_directory.joinpath("Metadata")
    #     def create_directories(self):
    #         self.root_directory.mkdir(exist_ok=True)

    class UserDirectories:
        def __init__(self) -> None:
            self.metadata_directory: Path | None = None
            self.download_directory: Path | None = None
            self.alt_download_directories: list[Path] = []
            self.legacy_download_directories: list[Path] = []
            self.legacy_metadata_directories: list[Path] = []

        def find_legacy_directory(
            self,
            directory_type: Literal["metadata", "download"] = "metadata",
            api_type: str = "",
        ):
            match directory_type:
                case "metadata":
                    directories = self.legacy_metadata_directories
                case _:
                    directories = self.legacy_download_directories
            final_directory = directories[0]
            for directory in directories:
                for part in directory.parts:
                    if api_type in part:
                        return directory
            return final_directory

    async def walk(self, directory: Path):
        return await usc_helper.walk(directory)


class FileManager:
    def __init__(self, directory_manager: DirectoryManager) -> None:
        self.files: list[Path] = []
        self.directory_manager = directory_manager

    async def set_default_files(
        self,
    ):
        assert self.directory_manager.user.metadata_directory
        assert self.directory_manager.user.download_directory
        await self.update_files(self.directory_manager.user.metadata_directory)
        await self.update_files(self.directory_manager.user.download_directory)

    async def refresh_files(self):
        return await self.set_default_files()

    async def update_files(self, directory: Path):
        directory_manager = self.directory_manager
        files = await directory_manager.walk(directory)
        self.files.extend(files)
        return files

    def add_file(self, filepath: Path):
        self.files.append(filepath)
        return True

    def remove_file(self, filepath: Path):
        if filepath in self.files:
            self.files.remove(filepath)
            return True
        return False

    def rename_path(self, old_filepath: Path, new_filepath: Path):
        self.remove_file(old_filepath)
        self.add_file(new_filepath)
        new_filepath.parent.mkdir(exist_ok=True, parents=True)
        shutil.move(old_filepath, new_filepath)
        return True

    def delete_path(self, filepath: Path):
        if filepath.is_dir():
            filepath.rmdir()
        else:
            self.remove_file(filepath)
            filepath.unlink(missing_ok=True)
        return True

    async def cleanup(self):
        unique: set[Path] = set()
        await self.refresh_files()
        for valid_file in self.find_string_in_path("__drm__"):
            self.delete_path(valid_file)
            unique.add(valid_file.parent)
        for unique_file in unique:
            self.delete_path(unique_file)
        return True

    def find_string_in_path(self, string: str):
        valid_files: list[Path] = []
        for file in self.files:
            if string in file.as_posix():
                valid_files.append(file)
        return valid_files

    async def find_metadata_files(self, legacy_files: bool = True):
        new_list: list[Path] = []
        for filepath in self.files:
            if not legacy_files:
                if "__legacy_metadata__" in filepath.parts:
                    continue
            match filepath.suffix:
                case ".db":
                    red_list = ["thumbs.db"]
                    status = [x for x in red_list if x == filepath.name.lower()]
                    if status:
                        continue
                    new_list.append(filepath)
                case ".json":
                    new_list.append(filepath)
                case _:
                    pass
        return new_list

    async def merge_alternative_directories(self, alt_directories: list[Path]):
        directory_manager = self.directory_manager
        assert directory_manager.user.download_directory
        for alt_download_directory in alt_directories:
            alt_files = await directory_manager.walk(alt_download_directory)
            for alt_file in alt_files:
                new_filepath = Path(
                    alt_file.as_posix().replace(
                        alt_download_directory.as_posix(),
                        directory_manager.user.download_directory.as_posix(),
                    )
                )

                if alt_file.suffix in [".json", ".db"]:
                    if new_filepath.exists():
                        new_filepath = usc_helper.find_unused_filename(new_filepath)
                        if new_filepath.exists():
                            breakpoint()
                if new_filepath.exists():
                    old_checksum = hashlib.md5(alt_file.read_bytes()).hexdigest()
                    new_checksum = hashlib.md5(new_filepath.read_bytes()).hexdigest()
                    if old_checksum == new_checksum:
                        self.delete_path(alt_file)
                    else:
                        old_size = alt_file.stat().st_size
                        new_size = new_filepath.stat().st_size
                        if old_size > new_size:
                            self.rename_path(alt_file, new_filepath)
                        elif new_size > old_size:
                            self.delete_path(alt_file)
                        elif old_size == new_size:
                            if usc_helper.is_image_valid(new_filepath):
                                self.delete_path(alt_file)
                            elif usc_helper.is_image_valid(alt_file):
                                self.rename_path(alt_file, new_filepath)
                else:
                    self.rename_path(alt_file, new_filepath)

            alt_files = await usc_helper.walk(alt_download_directory)
            if not alt_files:
                shutil.rmtree(alt_download_directory)


class FormatTypes:
    def __init__(self, site_settings: site_config_types) -> None:
        self.metadata_directory_format = site_settings.metadata_setup.directory_format
        self.file_directory_format = site_settings.download_setup.directory_format
        self.filename_format = site_settings.download_setup.filename_format

    def check_rules(self):
        """Checks for invalid filepath

        Returns:
            tuple(str,bool): Returns a string which explains invalid filepath format
        """
        bool_status = True
        wl = []
        invalid_list = []
        string = ""
        for key, _value in self:
            if key == "file_directory_format":
                bl = FormatAttributes()
                wl = [v for _k, v in bl.__dict__.items()]
                bl = bl.whitelist(wl)
                invalid_list = []
                for b in bl:
                    if b in self.file_directory_format.as_posix():
                        invalid_list.append(b)
            if key == "filename_format":
                bl = FormatAttributes()
                wl = [v for _k, v in bl.__dict__.items()]
                bl = bl.whitelist(wl)
                invalid_list = []
                for b in bl:
                    if b in self.filename_format.as_posix():
                        invalid_list.append(b)
            if key == "metadata_directory_format":
                wl = [
                    "{site_name}",
                    "{first_letter}",
                    "{model_id}",
                    "{profile_username}",
                    "{model_username}",
                ]
                bl = FormatAttributes().whitelist(wl)
                invalid_list: list[str] = []
                for b in bl:
                    if b in self.metadata_directory_format.as_posix():
                        invalid_list.append(b)
            if invalid_list:
                string += f"You cannot use {','.join(invalid_list)} in {key}. Use any from this list {','.join(wl)}"
                bool_status = False

        return string, bool_status

    def check_unique(self):
        values: list[str] = []
        unique = []
        new_format_copied = copy.deepcopy(self)
        option: dict[str, Any] = {}
        option["string"] = ""
        option["bool_status"] = True
        option["unique"] = new_format_copied
        f = FormatAttributes()
        for key, value in self:
            value: Path
            if key == "file_directory_format":
                unique = ["{media_id}", "{model_username}"]
                values = list(value.parts)
                option["unique"].file_directory_format = unique
            elif key == "filename_format":
                values = []
                unique = ["{media_id}", "{filename}"]
                for _key2, value2 in f:
                    if value2 in value.as_posix():
                        values.append(value2)
                option["unique"].filename_format = unique
            elif key == "metadata_directory_format":
                unique = ["{model_username}"]
                values = list(value.parts)
                option["unique"].metadata_directory_format = unique
            if key != "filename_format":
                e = [x for x in values if x in unique]
            else:
                e = [x for x in unique if x in values]
            if e:
                setattr(option["unique"], key, e)
            else:
                option[
                    "string"
                ] += f"{key} is a invalid format since it has no unique identifiers. Use any from this list {','.join(unique)}\n"
                option["bool_status"] = False
        return option

    def __iter__(self):
        for attr, value in self.__dict__.items():
            yield attr, value



from typing import Any

import aio_pika
import ujson


def create_notification(
    category: str,
    site_name: str,
    item: Any,
):
    json_message = {
        "site_name": site_name,
        "category": category,
        "performer_id": item.id,
        "username": item.username,
    }
    message = {"id": item.id, "data": json_message}
    return message


def create_message(
    site_name: str, item: Any, mandatory_jobs: dict[str, dict[str, list[str]]]
):
    json_message = {
        "site_name": site_name,
        "performer_id": item.id,
        "username": item.username,
        "mandatory_jobs": mandatory_jobs,
    }
    message = {"id": item.id, "data": json_message}
    return message


class AioPikaWrapper:
    def __init__(self, host: str = "localhost", heartbeat: int | None = None):
        heartbeat_param = f"?heartbeat={heartbeat}" if heartbeat is not None else ""
        self.amqp_url = f"amqp://{host}/{heartbeat_param}"
        self.connection = None
        self.channel = None

    def get_connection(self):
        assert self.connection is not None
        return self.connection

    def get_channel(self):
        assert self.channel is not None
        return self.channel

    async def connect(self, prefetch_count: int = 0):
        if self.connection and not self.connection.is_closed:
            return
        self.connection = await aio_pika.connect_robust(self.amqp_url)
        self.channel = await self.connection.channel()
        await self.channel.set_qos(prefetch_count=prefetch_count)

    async def disconnect(self):
        if self.connection:
            await self.connection.close()

    async def declare_queue(self, queue_name: str, durable: bool = True):
        if not self.channel or self.channel.is_closed:
            await self.connect()
        assert self.channel is not None
        return await self.channel.declare_queue(
            queue_name,
            durable=durable,
            arguments={"x-message-deduplication": True, "x-max-priority": 10},
        )

    async def publish_message(
        self,
        queue_name: str,
        message: dict[str, Any],
        durable: bool = True,
        priority: int | None = None,
        print_error: bool = True,
    ):
        if self.channel is None:
            await self.connect()
        await self.declare_queue(queue_name, durable)
        assert self.channel is not None

        message_id = message.get(
            "id", "default_id"
        )  # Ensure you have a unique ID for deduplication
        headers = {"x-deduplication-header": message_id}
        try:
            await self.channel.default_exchange.publish(
                aio_pika.Message(
                    body=ujson.dumps(message).encode(),
                    delivery_mode=(
                        aio_pika.DeliveryMode.PERSISTENT
                        if durable
                        else aio_pika.DeliveryMode.NOT_PERSISTENT
                    ),
                    headers=headers,
                    priority=priority,
                ),
                routing_key=queue_name,
            )
            print(f"Message published to {queue_name}")
            return True
        except aio_pika.exceptions.DeliveryError as e:
            if print_error:
                print(f"Error publishing message: {e}")
            return False

    async def publish_notification(self, message: dict[str, Any]):
        await self.publish_message("telegram_notifications", message)
        await self.publish_message("discord_notifications", message)

    async def push_back_message(
        self,
        queue_name: str,
        message: aio_pika.abc.AbstractIncomingMessage,
        priority: int | None = None,
    ):
        await message.reject()
        task = ujson.loads(message.body.decode())
        await self.publish_message(queue_name, task, priority=priority)

    async def close(self):
        if self.connection:
            await self.connection.close()



from ultima_scraper_collection.config import UltimaScraperCollectionConfig

config = UltimaScraperCollectionConfig().load_default_config()
for directory_path in config.site_apis.onlyfans.download_setup.directories:
    assert directory_path.path
    directory_path = directory_path.path.joinpath("OnlyFans")
    if not directory_path.exists():
        continue
    sorted_dirs = sorted(directory_path.iterdir(), key=lambda x: x.name.lower())
    total_bytes_deleted = 0
    for letter in sorted_dirs:
        print(letter)
        letter_path = directory_path.joinpath(letter)
        for performer_folder in letter_path.iterdir():
            messages_folder = performer_folder.joinpath("Messages")
            mass_messages_folder = performer_folder.joinpath("MassMessages")
            if messages_folder.exists() and mass_messages_folder.exists():
                free_messages_folder = messages_folder.joinpath("Free")
                free_mass_messages_folder = mass_messages_folder.joinpath("Free")
                paid_messages_folder = messages_folder.joinpath("Paid")
                paid_mass_messages_folder = mass_messages_folder.joinpath("Paid")
                collection = [
                    [free_messages_folder, free_mass_messages_folder],
                    [paid_messages_folder, paid_mass_messages_folder],
                ]
                for value_message_folder, value_mass_message_folder in collection:
                    if (
                        value_message_folder.exists()
                        and value_mass_message_folder.exists()
                    ):
                        v_m_f = value_message_folder.rglob("*.mp4")
                        v_m_m_f = value_mass_message_folder.rglob("*.mp4")
                        for video_filepath in v_m_m_f:
                            for video_filepath2 in v_m_f:
                                if video_filepath.name == video_filepath2.name:
                                    video_filepath_bytes = video_filepath.read_bytes()
                                    video_filepath2_bytes = video_filepath2.read_bytes()
                                    if video_filepath_bytes == video_filepath2_bytes:
                                        print(
                                            f"Duplicate found: {video_filepath} and {video_filepath2}"
                                        )
                                        video_filepath2.unlink()
                                        total_bytes_deleted += (
                                            video_filepath.stat().st_size
                                        )
                                        print(
                                            f"Total bytes deleted: {total_bytes_deleted / (1024 * 1024 * 1024):.2f} GB"
                                        )
                                        pass
                                    else:
                                        print(
                                            f"Intervention needed: {video_filepath} and {video_filepath2}"
                                        )
                                        input("Press enter to continue.")



from typing import Any, Callable, Coroutine

import aio_pika
import orjson


class MandatoryJob:
    def __init__(
        self,
        job_name: str,
        success_queue_name: str | None = None,
        failure_queue_name: str | None = None,
    ) -> None:
        self.job_name = job_name
        self.success_queue_name = success_queue_name
        self.failure_queue_name = failure_queue_name

    def to_dict(self) -> dict[str, Any]:
        return {
            "job_name": self.job_name,
            "success_queue_name": self.success_queue_name,
            "failure_queue_name": self.failure_queue_name,
        }

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> "MandatoryJob":
        return cls(**data)


class StandardData:
    def __init__(
        self,
        site_name: str,
        performer_id: int,
        performer_username: str,
        category: str = "",
        mandatory_jobs: list[MandatoryJob] = [],
        extra_data: dict[str, Any] = {},
    ) -> None:
        self.site_name = site_name
        self.performer_id = performer_id
        self.performer_username = performer_username
        self.mandatory_jobs = mandatory_jobs
        self.category = category
        self.extra_data = extra_data

    def to_dict(self) -> dict[str, Any]:
        return {
            "site_name": self.site_name,
            "performer_id": self.performer_id,
            "performer_username": self.performer_username,
            "mandatory_jobs": [x.to_dict() for x in self.mandatory_jobs],
            "category": self.category,
            "extra_data": self.extra_data,
        }

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> "StandardData":
        mandatory_jobs = [
            MandatoryJob.from_dict(x) for x in data.get("mandatory_jobs", [])
        ]
        data.pop("mandatory_jobs", None)
        return cls(**data, mandatory_jobs=mandatory_jobs)

    @classmethod
    def from_legacy(cls, data: dict[str, Any]) -> "StandardData":
        temp_data = data["data"]
        return cls(
            site_name=temp_data["site_name"],
            performer_id=temp_data["performer_id"],
            performer_username=temp_data["username"],
        )

    def create_mandatory_job(
        self,
        job_name: str,
        success_queue_name: str | None = None,
        failure_queue_name: str | None = None,
    ) -> MandatoryJob:
        # Check if a job with the same job_name already exists
        for job in self.mandatory_jobs:
            if job.job_name == job_name:
                return job  # Return the existing job if found

        # If no match is found, create and add the new job
        mandatory_job = MandatoryJob(job_name, success_queue_name, failure_queue_name)
        self.mandatory_jobs.append(mandatory_job)
        return mandatory_job


class UltimaQueue:
    def __init__(self, host: str = "localhost") -> None:
        self.amqp_url = f"amqp://{host}"
        self.connection = None
        self.channel = None

    def get_connection(self):
        assert self.connection is not None
        return self.connection

    def get_channel(self):
        assert self.channel is not None
        return self.channel

    async def connect(self, prefetch_count: int = 0):
        if self.connection and not self.connection.is_closed:
            return
        self.connection = await aio_pika.connect_robust(self.amqp_url)
        self.channel = await self.connection.channel()
        await self.channel.set_qos(prefetch_count=prefetch_count)

    async def disconnect(self):
        if self.connection:
            await self.connection.close()

    async def declare_queue(self, queue_name: str, durable: bool = True):
        if not self.channel or self.channel.is_closed:
            await self.connect()
        assert self.channel is not None
        return await self.channel.declare_queue(
            queue_name,
            durable=durable,
            arguments={"x-message-deduplication": True, "x-max-priority": 10},
        )

    async def consume_messages(
        self,
        queue_name: str,
        callback: Callable[[Any, Any], Coroutine[Any, Any, None]],
        *args: Any,
        prefetch_count: int = 0,
    ):
        """Consume messages from a specified queue."""
        if not self.channel or self.channel.is_closed:
            await self.connect(prefetch_count=prefetch_count)
        queue = await self.declare_queue(queue_name)

        async with queue.iterator() as queue_iter:
            async for message in queue_iter:
                try:
                    async with message.process(requeue=True, ignore_processed=True):
                        await callback(message, *args)
                        # await message.ack()
                except Exception as _e:
                    continue
        await self.disconnect()

    async def publish_message(
        self,
        queue_name: str,
        message: dict[str, Any],
        durable: bool = True,
        priority: int | None = None,
        unique_id: int | str | None = None,
        headers: dict[str, Any] = {},
        print_error: bool = True,
        surpress: bool = False,
    ):
        if self.channel is None:
            await self.connect()
        await self.declare_queue(queue_name, durable)
        assert self.channel is not None

        if unique_id is None:
            unique_id = message.get("id")
        if unique_id:
            # unique_id = str(unique_id)
            if not headers:
                headers = {"x-deduplication-header": unique_id}
        try:
            await self.channel.default_exchange.publish(
                aio_pika.Message(
                    body=orjson.dumps(message, option=orjson.OPT_NON_STR_KEYS),
                    delivery_mode=(
                        aio_pika.DeliveryMode.PERSISTENT
                        if durable
                        else aio_pika.DeliveryMode.NOT_PERSISTENT
                    ),
                    headers=headers,
                    priority=priority,
                ),
                routing_key=queue_name,
            )
            if not surpress:
                print(f"Message published to {queue_name}")
            return True
        except aio_pika.exceptions.DeliveryError as e:
            if print_error:
                print(f"Error publishing message: {e}")
            return False

    async def publish_notification(self, message: dict[str, Any]):
        await self.publish_message("telegram_notifications", message)
        await self.publish_message("discord_notifications", message)

    async def push_back_message(
        self,
        queue_name: str,
        message: aio_pika.abc.AbstractIncomingMessage,
        priority: int | None = None,
    ):
        await message.reject()
        task = orjson.loads(message.body.decode())
        await self.publish_message(
            queue_name, task, priority=priority, headers=message.headers
        )

    async def close(self):
        if self.connection:
            await self.connection.close()



import asyncio
import copy
import traceback
from pathlib import Path
from urllib.parse import urlparse

import ffmpeg
from aiohttp import ClientConnectionError, ClientResponse
from alive_progress import alive_bar
from ultima_scraper_api import auth_types
from ultima_scraper_api.apis.onlyfans.classes.mass_message_model import MassMessageModel
from ultima_scraper_api.helpers import main_helper
from ultima_scraper_collection.managers.database_manager.connections.sqlite.models.media_model import (
    TemplateMediaModel,
)
from ultima_scraper_collection.managers.filesystem_manager import FilesystemManager
from ultima_scraper_collection.managers.metadata_manager.metadata_manager import (
    MediaMetadata,
    invalid_subdomains,
)
from ultima_scraper_db.databases.ultima_archive.schemas.templates.site import (
    MediaModel,
    MessageModel,
)
from ultima_scraper_renamer.reformat import ReformatManager


class DownloadManager:
    def __init__(
        self,
        authed: auth_types,
        filesystem_manager: FilesystemManager,
        media_set: set[MediaMetadata] = set(),
        reformat: bool = True,
    ) -> None:
        self.authed = authed
        self.filesystem_manager = filesystem_manager
        self.auth_session = self.authed.auth_session
        self.requester = self.authed.get_requester()
        self.content_list: set[MediaMetadata] = media_set
        self.errors: list[TemplateMediaModel] = []
        self.reformat = reformat
        self.reformat_manager = ReformatManager(self.authed, filesystem_manager)
        self.bar = None

    async def check_total_size(self):
        async def get_file_size(file: MediaMetadata):
            if len(file.urls) > 1:
                return 0  # Skip files with multiple URLs as in your logic
            response = await self.auth_session.active_session.head(file.urls[0])
            return response.content_length if response.content_length else 0

        tasks = [get_file_size(file) for file in self.content_list]
        sizes = await asyncio.gather(*tasks)
        return sum(sizes)

    async def process_download_item(self, download_item: MediaMetadata):
        try:
            db_media = download_item.__db_media__
        except Exception as _e:
            pass
        assert db_media
        content = download_item.get_content_metadata()
        if content:
            db_content = content.__db_content__
            assert db_content
            if isinstance(db_content, MessageModel):
                if db_content.queue_id:
                    db_filepath = db_media.find_filepath(
                        (db_content.queue_id, "MassMessages")
                    )
                else:
                    db_filepath = db_media.find_filepath()
            else:
                db_filepath = db_media.find_filepath()
        else:
            db_filepath = db_media.find_filepath()
        if isinstance(download_item, MediaMetadata):
            final_filepath = download_item.directory.joinpath(download_item.filename)
            if final_filepath.exists():
                final_size = final_filepath.stat().st_size
                timestamp = db_media.created_at.timestamp()
                await main_helper.format_file(
                    final_filepath,
                    timestamp,
                    self.reformat,
                )
                if db_media and db_filepath:
                    if not db_filepath.preview:
                        db_media.size = download_item.size = final_size
                    else:
                        if final_size > db_media.size:
                            db_media.size = final_size
                    db_filepath.downloaded = True

    async def bulk_download(self):
        queue: asyncio.Queue[tuple[Path, list[Path]]] = asyncio.Queue()
        worker_task = asyncio.create_task(self.ffmpeg_worker(queue))
        final_list = [
            self.download(media_item, queue) for media_item in self.content_list
        ]
        global_settings = self.authed.get_api().get_global_settings()
        decrypt_media_path = global_settings.drm.decrypt_media_path
        if final_list:
            with alive_bar(
                len(self.content_list),
                title="Downloading Media Items",
                spinner="dots",
                bar="smooth",
                enrich_print=True,
            ) as bar:
                self.bar = bar  # Allow updates within `self.download` if needed

                # Run tasks asynchronously and track progress
                results = await asyncio.gather(*final_list, return_exceptions=True)

            drm_results = [
                x for x in results if x and isinstance(x, MediaMetadata) and x.drm
            ]
            if drm_results:
                with alive_bar(
                    len(drm_results),
                    title="Processing DRM-protected files",
                    spinner="dots",
                    bar="smooth",
                    enrich_print=True,
                ) as bar:
                    authed_drm = self.authed.drm
                    for download_item in drm_results:
                        try:
                            final_download_path = download_item.directory.joinpath(
                                download_item.filename
                            )

                            if authed_drm and download_item.drm:
                                bar.text(f"Decrypting: {download_item.filename}")
                                decrypted_media_paths: list[Path] = []
                                for download_path in download_item.drm_download_paths:
                                    output_filepath = authed_drm.decrypt_file(
                                        download_path,
                                        download_item.key,
                                        temp_output_path=decrypt_media_path,
                                    )
                                    if not output_filepath:
                                        raise Exception("No output_filepath")
                                    decrypted_media_paths.append(output_filepath)

                                bar.text(f"Merging: {download_item.filename}")
                                future = asyncio.get_event_loop().create_future()
                                task = (
                                    final_download_path,
                                    decrypted_media_paths,
                                    future,
                                )
                                await queue.put(task)
                                await future
                            bar()
                        except Exception as e:
                            pass

            tasks = []
            for download_item in results:
                if not download_item:
                    continue
                tasks.append(self.process_download_item(download_item))
            await asyncio.gather(*tasks)
        await queue.join()
        worker_task.cancel()
        await asyncio.gather(worker_task, return_exceptions=True)
        pass

    async def drm_download(self, download_item: MediaMetadata):
        content_metadata = download_item.__content_metadata__
        authed = self.authed
        reformat_manager = ReformatManager(authed, self.filesystem_manager)
        assert reformat_manager.filesystem_manager.directory_manager
        site_config = reformat_manager.filesystem_manager.directory_manager.site_config
        drm = authed.drm
        media_item = download_item.__raw__
        assert drm and media_item
        mpd = await drm.get_mpd(media_item)
        pssh = await drm.get_pssh(mpd)
        responses: list[ClientResponse] = []

        if pssh:
            if content_metadata:
                soft_data = content_metadata.__soft__
                raw_data = soft_data.__raw__.copy()
                if (
                    isinstance(soft_data, MassMessageModel)
                    and soft_data
                    and soft_data.author.is_authed_user()
                ):
                    raw_data["responseType"] = ""
            else:
                raw_data = {"responseType": ""}
            license = await drm.get_license(raw_data, media_item, pssh)
            keys = await drm.get_keys(license)
            content_key = keys[-1]
            key = f"{content_key.kid.hex}:{content_key.key.hex()}"
            download_item.key = key
            video_url, audio_url = [
                drm.get_video_url(mpd, media_item),
                drm.get_audio_url(mpd, media_item),
            ]
            download_item.urls = [video_url]
            reformat_item = reformat_manager.prepare_reformat(download_item)
            file_directory = reformat_item.reformat(
                site_config.download_setup.directory_format
            )
            reformat_item.directory = file_directory
            file_path = reformat_item.reformat(
                site_config.download_setup.filename_format
            )
            download_item.directory = file_directory
            download_item.filename = file_path.name
            for media_url in video_url, audio_url:
                drm_download_item = copy.copy(download_item)
                drm_download_item = reformat_manager.drm_format(
                    media_url, drm_download_item
                )

                signature_str = await drm.get_signature(media_item)
                response = await authed.auth_session.request(
                    media_url, premade_settings="", custom_cookies=signature_str
                )
                responses.append(response)
        return responses

    async def download(
        self,
        download_item: MediaMetadata,
        ffmpeg_queue: asyncio.Queue[tuple[Path, list[Path]]],
    ):
        try:
            if not download_item.urls:
                return
            attempt = 0
            try:
                db_media = download_item.__db_media__
            except Exception as _e:
                pass
            assert db_media
            await db_media.awaitable_attrs.content_media_assos
            p_url = urlparse(download_item.urls[0])
            assert p_url.hostname
            subdomain = p_url.hostname.split(".")[0]
            if any(subdomain in nm for nm in invalid_subdomains):
                return

            authed = self.authed
            authed_drm = authed.drm

            async with self.auth_session.semaphore:
                while (
                    attempt < self.auth_session.get_session_manager().max_attempts + 1
                ):
                    try:
                        if download_item.drm:

                            if not authed_drm:
                                break
                            responses = await self.drm_download(download_item)
                        else:
                            responses = [
                                await self.requester.request(download_item.urls[0])
                            ]
                        if all(response.status != 200 for response in responses):
                            attempt += 1
                            continue
                        if not download_item.directory:
                            raise Exception(
                                f"{download_item.id} has no directory\n {download_item}"
                            )
                        final_size = 0
                        error = None
                        process_responses = True
                        if download_item.drm and await self.drm_check_downloaded(
                            download_item
                        ):
                            process_responses = False
                        if process_responses:
                            for response in responses:
                                download_path, error = await self.writer(
                                    response,
                                    download_item,
                                    encrypted=bool(download_item.key),
                                )
                                if error:
                                    attempt += 1
                                    break
                                if download_item.drm and download_path:
                                    if (
                                        download_path
                                        not in download_item.drm_download_paths
                                    ):
                                        download_item.drm_download_paths.append(
                                            download_path
                                        )
                                # if authed_drm and download_item.drm and download_path:
                                #     output_filepath = authed_drm.decrypt_file(
                                #         download_path, download_item.key
                                #     )
                                #     if not output_filepath:
                                #         raise Exception("No output_filepath")
                                #     decrypted_media_paths.append(output_filepath)
                                if response.content_length:
                                    final_size += response.content_length
                        if error == 1:
                            # Server Disconnect Error
                            continue
                        elif error == 2:
                            # Resource Not Found Error
                            break
                        assert download_item.filename
                        download_path = download_item.directory.joinpath(
                            download_item.filename
                        )
                        break
                    except asyncio.TimeoutError as _e:
                        continue
                    except ClientConnectionError as _e:
                        attempt += 1
                        continue
                    except Exception as _e:
                        traceback.print_exc()
            self.bar()
            return download_item
        except Exception as _e:
            pass

    async def writer(
        self,
        result: ClientResponse,
        download_item: MediaMetadata,
        encrypted: bool = True,
    ):
        async with result as response:
            if download_item.drm and encrypted:
                download_item = copy.copy(download_item)
                download_item = self.reformat_manager.drm_format(
                    response.url.human_repr(), download_item
                )
                downloaded = await self.drm_check_downloaded(download_item)
                if downloaded:
                    return download_item.get_filepath(), None
            assert download_item.directory and download_item.filename
            download_path = Path(download_item.directory, download_item.filename)
            db_media = copy.copy(download_item.__db_media__)
            db_media.directory = download_item.directory
            db_media.filename = download_item.filename
            download = await self.check(db_media, response)
            if not download:
                return download_path, None
            failed = await self.filesystem_manager.write_data(response, download_path)
            return download_path, failed

    async def drm_check_downloaded(self, download_item: MediaMetadata):
        download_path = download_item.get_filepath()
        if ".enc" in download_path.name:
            download_path = Path(download_path.as_posix().replace(".enc", ".dec"))
        if download_path.exists():
            try:
                ffmpeg.probe(download_path.as_posix())
                return True
            except ffmpeg.Error as _e:
                return False
        return False

    async def check(self, download_item: MediaModel, response: ClientResponse):
        # Checks if we should download item or not // True | False
        filepath = Path(download_item.directory, download_item.filename)
        response_status = False
        if response.status == 200:
            response_status = True
            if response.content_length:
                download_item.size = response.content_length

        if filepath.exists():
            try:
                if filepath.stat().st_size == response.content_length:
                    return False
                else:
                    return True
            except Exception as _e:
                pass
        else:
            if response_status:
                # Can produce false positives due to the same reason below
                return True
            else:
                # Reached this point because it probably exists in the folder but under a different content category
                pass

    def get_format(self, filepath: Path) -> str:
        """Get the container format of a media file using ffprobe."""
        try:
            probe = ffmpeg.probe(filepath.as_posix())
            return probe["format"][
                "format_name"
            ]  # Short format name (e.g., 'mp4', 'mov')
        except ffmpeg.Error as e:
            raise Exception(f"Error probing file {filepath}: {e.stderr.decode()}")

    def get_preferred_format(self, filepath: Path) -> str:
        """Extract the preferred container format from the file."""
        try:
            probe = ffmpeg.probe(filepath.as_posix())
            format_names = probe["format"][
                "format_name"
            ]  # E.g., "mov,mp4,m4a,3gp,3g2,mj2"
            formats = format_names.split(",")  # Split into a list
            # Prioritize 'mp4' if available, otherwise take the first format
            return "mp4" if "mp4" in formats else formats[0]
        except ffmpeg.Error as e:
            raise Exception(f"Error probing file {filepath}: {e.stderr.decode()}")

    async def format_media(
        self, output_filepath: Path, decrypted_media_paths: list[Path]
    ):
        # If you have decrypted video and audio to merge
        if len(decrypted_media_paths) > 1:
            dec_video_path, dec_audio_path = decrypted_media_paths
            video_input = ffmpeg.input(dec_video_path)  # type:ignore
            audio_input = ffmpeg.input(dec_audio_path)  # type:ignore

            try:
                # Dynamically determine the preferred format
                output_format = self.get_preferred_format(dec_video_path)

                temp_output_filepath = output_filepath.with_suffix(".part")
                _ffmpeg_output = ffmpeg.output(  # type:ignore
                    video_input,  # type:ignore
                    audio_input,  # type:ignore
                    temp_output_filepath.as_posix(),
                    vcodec="copy",
                    acodec="copy",
                    f=output_format,  # Dynamically set the format
                ).run(capture_stdout=True, capture_stderr=True, overwrite_output=True)

                temp_output_filepath.rename(output_filepath)
                return True
            except ffmpeg.Error as _e:
                return False
        return True

    async def ffmpeg_worker(self, queue: asyncio.Queue[tuple[Path, list[Path]]]):
        while True:
            output_filepath, decrypted_media_paths, future = await queue.get()
            try:
                result = await self.format_media(output_filepath, decrypted_media_paths)
                future.set_result(result)
                for path in decrypted_media_paths:
                    path.unlink()
            except Exception as e:
                future.set_exception(e)
            finally:
                queue.task_done()
